/**
  * Redistribution and use of this software and associated
  * documentation ("Software"), with or without modification, are
  * permitted provided that the following conditions are met:
  *
  * 1. Redistributions of source code must retain copyright statements
  *    and notices.  Redistributions must also contain a copy of this
  *    document.
  *
  * 2. Redistributions in binary form must reproduce the above
  *    copyright notice, this list of conditions and the following
  *    disclaimer in the documentation and/or other materials provided
  *    with the distribution.
  *
  * 3. The names "TRANSORG" or "EUTIST-AMI" must not be used to endorse or promote
  *    products derived from this Software without prior written
  *    permission of one of the parties involved in the TRANSORG project. For
  * 	  written permission, please contact diasz@ida.liu.se.
  *
  * 4. Products derived from this Software may not be called "TRANSORG"
  *    nor may "TRANSORG" appear in their names without prior written
  *    permission.
  *
  * 5. Due credit should be given to the TRANSORG Project
  *    (http://www.ida.liu.se/~rtslab/TRANSORG/).
  *
  * THIS SOFTWARE IS PROVIDED BY THE TRANSORG PROJECT CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
  * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  * PARTICULAR PURPOSE ARE DISCLAIMED.
  *
  * IN NO EVENT SHALL THE TRANSORG PROJECT CONTRIBUTORS
  * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
  * Copyright 2003 (C) TRANSORG Project Contributors All Rights Reserved.
  *
  * $Id: SampleCompiler.java,v 1.0   Exp $
  *
  */

/** Author		; Diana Szentivanyi
* Modified		: Isabelle Ravot
* Last changed	: 24/04/2003
* Contact		: diasz@ida.liu.se
*/

/** Sample Compiler class used to process trace information written in the appropriate files while requests are travelling from client to server and back
  * The idea is to take those files, in which information about starting and ending of certain phases in the request travelling are traced and parse them.
  * The informations writtne there are of the form request client identifier, request retention identifier, moment(phase) code and start time for that phase.
  * For example, one moment is the entering of the server request in cold passive replication: thus, a record written in such a file can contain:
  *  2(constant coresponding to ENTER_RECEIVE_REQUEST)
  *  234 (Retention identifier of the request)
  *  0*ORB@123lkdiw8222 (Client identifier of the request)
  *  124985436578435 (the start time of the phase, in milliseconds since jan 1st 1970 - as returned by System.currentTimeMillis())
  * All trace files are read and these types of records are collected and inserted in a hashtable containing as keys the unique request identifiers (client_id+
  * retention_id). A vector containing all moment records for a certain request is sorted according to time. The data about requests (such as how long a certain
  * operation lasted - e.g. how long the request spent in the server interceptor) are written in files corresponding to every method name, and request phase name.
  * For example, if a method with name "add_numbers" was called 100 times during the experiment then there are 100 values to be written in the file add_numbers_SCL,
  * 100 values to be written in the file add_numbers_TTS, 100 values to be written in the file add_numbers_TC, etc. The file add_numbers_SCL contains time intervals
  * that the request spent in the client interceptor before travelling to the server, in all the hundred occasions when the method add_numbers was called. Of course,
  * the 100 values correspond to 100 requests with different retention identifiers and perhaps the same client identifier, but to the same name of the method that was
  * called. There is also a file corresponding to total roundtrip times. The results that are provided as statistics are obtained by averaging the e.g. 100 values from the
  * files.
  * NOTE: the implementation of the Compiler can be changed so that it fits trace processing needs, as desired. For, example, other traces can be introduced
  * and these need other processing steps than present in the Compiler's present implementation
   * @author <a href="mailto:diasz@ida.liu.se"> Diana Szentivanyi </a>
   * @version $Revision: 1.0 $ $Date:2003/06/23$
 */

public class Compiler{

    final static int NR_CHARS_OPER=8;
    final static int NR_DIGITS_NR=5;


    private class ReqTraceStr{

	int moment_code;
	int retention_id;
	String client_id;
	long time;
	String oper_name;


	public ReqTraceStr(int moment_code,int retention_id,String client_id,long time,String oper_name){
	    this.moment_code=moment_code;
	    this.time=time;
	    this.client_id=client_id;
	    this.retention_id=retention_id;
	    this.oper_name=oper_name;
	}
    }

    private class GetStateTrStr{
	int moment_code;
	int size;
	long time;

	public GetStateTrStr(int moment_code,int size,long time){
	    this.moment_code=moment_code;
	    this.size=size;
	    this.time=time;
	}
    }

    private class AvRecStr{

	long sum,sum_without_wait,sum_sq,sum_sq_without_wait,max_value,max_value_without_wait;

	int nr_of_calls;
	java.util.Vector values;

	public AvRecStr(){
	    sum=0;
	    sum_without_wait=0;
	    sum_sq=0;
	    sum_sq_without_wait=0;
	    nr_of_calls=0;
	    max_value=-1;
	    max_value_without_wait=-1;
	    values=new java.util.Vector();
	}
    }

    private class AvWaitRecStr{
	long sum;
	int nr_of_calls;

	public AvWaitRecStr(){
	    sum=0;
	    nr_of_calls=0;
	}
    }

    private class AvSliceRecStr{
	long sum;
	int nr_of_calls;
	String oper;
	long sum_sq;
	long max_value;

	public AvSliceRecStr(String oper){
	    nr_of_calls=0;
	    sum=0;
	    sum_sq=0;
	    max_value=-1;
	    this.oper=oper;
	}
    }


    private void read_req_t_line(java.io.RandomAccessFile raf){

	int moment_code;
	int retention_id;
	String client_id,hashcode,oper_name;
	long time;
	boolean put=false;
	try{
	    moment_code=raf.readInt();

	    retention_id=raf.readInt();
	    time=raf.readLong();
	    client_id=raf.readLine();
	    oper_name=raf.readLine();

	    //System.out.println("REad "+oper_name+" "+retention_id+" "+moment_code);
	    hashcode=client_id+"+"+oper_name+"+"+retention_id;
	    java.util.Vector req_trace_v=(java.util.Vector)req_trace_t.get(hashcode);
	    if(req_trace_v==null){
		put=true;
		req_trace_v=new java.util.Vector();
	    }
	    //if(req_trace_v.size()>36) return;
	    if(oper_name==null) System.out.println("oopps opername is null "+retention_id);
	    req_trace_v.addElement(new ReqTraceStr(moment_code,retention_id,client_id,time,oper_name));
	    if(put){
		req_trace_t.put(hashcode,req_trace_v);
	    }

	}catch(java.lang.Exception e){e.printStackTrace();}
    }

    private void switch_elements(java.util.Vector v,java.lang.Object obj,int pos){
	v.addElement(obj);
	v.remove(pos);
    }

    private boolean is_not_ok(int l,boolean[] b_v_s){
	int i;
	boolean ret=false;
	for(i=0;i<l;i++)
	    ret=ret||!b_v_s[i];
	return ret;
    }

    private AvSliceRecStr get_slice(String oper,java.util.Vector slices){
	int l=slices.size(),i;
	AvSliceRecStr asrs;
	for(i=2;i<l;i++){
	    asrs=(AvSliceRecStr)slices.elementAt(i);
	    if(asrs.oper.equals(oper))
		return asrs;
	}
	asrs=new AvSliceRecStr(oper);
	slices.addElement(asrs);
	return asrs;
    }

    private void print_vector(java.util.Vector v){
	int l=v.size(),i;

	for(i=0;i<l;i++)
	    System.out.print(((ReqTraceStr)v.elementAt(i)).moment_code+" "+((ReqTraceStr)v.elementAt(i)).time+"  ");
	System.out.println();
    }



    private String pad_str(String s,int nl){
	int i,l;
	String toRet="";
	l=s.length();
	if(l==nl) return s;
	else{
	    for(i=0;i<(nl-l)/2;i++)
		toRet+=" ";
	    toRet+=s;
	    for(i=0;i<nl-l-(nl-l)/2;i++)
		toRet+=" ";
	    return toRet;
	}
    }

    private String pad_nr(int nr,int nl){

	String toRet=Integer.toString(nr);

	if(toRet.length()==nl) return toRet;
	StringBuffer toRetf=new StringBuffer();
	for(int i=0;i<nl-toRet.length();i++)
	    toRetf.append(" ");
	toRetf.append(toRet);
	return toRetf.toString();
    }

    private void sort_g_s_vector_t(java.util.Vector v){
	int i,l,j;
	GetStateTrStr crt,next;

	l=v.size();
	for(i=0;i<l;i++)
	    for(j=0;j<l-i-1;j++){
		crt=(GetStateTrStr)v.elementAt(j);
		next=(GetStateTrStr)v.elementAt(j+1);
		if(crt.time>next.time){
		    if(j<l-2){
			v.insertElementAt(crt,j+2);
			v.remove(j);
		    }
		    else
			{
			    v.insertElementAt(next,j);
			    v.remove(l);
			}
		}
	    }
    }


    private void sort_vector_t(java.util.Vector v){
	int i,l,j;
	ReqTraceStr crt,next;

	l=v.size();
	for(i=0;i<l;i++)
	    for(j=0;j<l-i-1;j++){
		crt=(ReqTraceStr)v.elementAt(j);
		next=(ReqTraceStr)v.elementAt(j+1);
		if(crt.time>next.time){
		    if(j<l-2){
			v.insertElementAt(crt,j+2);
			v.remove(j);
		    }
		    else
			{
			    v.insertElementAt(next,j);
			    v.remove(l);
			}
		}
	    }
    }

    public Compiler(){
	repl_m_trace_t=new java.util.Hashtable();
	req_trace_t=new java.util.Hashtable();
	times=new java.util.Hashtable();
	oper_av_t=new java.util.Hashtable();
	oper_av_w_t=new java.util.Hashtable();
	get_state_trace_t=new java.util.Hashtable();
    }

    java.util.Hashtable repl_m_trace_t;
    java.util.Hashtable req_trace_t,times;
    static java.util.Hashtable oper_av_t,oper_av_w_t,get_state_trace_t;
    static int current_repl_style;


    public void parse_req_trace(String f_name){
	java.io.RandomAccessFile req_trace;
	boolean error=false;

	try{
	    req_trace=new java.io.RandomAccessFile(f_name,"rw");

	    long length=req_trace.length();

	    while(req_trace.getFilePointer()<length){

		read_req_t_line(req_trace);
	    }
	    req_trace.close();
	}catch(java.lang.Exception e){}
    }

    private String spaces(int nr){
	String t_r="";
	int i;
	for(i=0;i<nr;i++)
	    t_r+=" ";
	return t_r;
    }


    public void compute_req_times(String experim_name,String value_file_name,String value_file_w_wait_name){
	/*

	   SCL - stay in client interceptor
	   TTS - travel to server
           SSL - stay in the serverinterceptor rigt leader
	   SSNL - stay in the serverinterceptor whithout sending back the reply because the server wasn't the leader
	   SRA - stay in the serverinterceptor when the reply was already there
	   TPSNL - time to process the request when the server is not the leader
	   TPSL  - time to process the request when the server is the leader
	   TPRA - time to process the request when the reply was already there
	   TW - time to wait before started to treat the request
	   TWRAC - time to watch if the request has already been computed

	   TGSB - time to get the state before the execution
	   TER - time to execute the request
	   TGSA - time to get the state after the execution
	   TC - time to do the consensus
	   TRS - time to restore the state in case the server is not the leader
	   SETS - time to restore the state when the server was the leader but it's not the request it has proposed that has been chosen
	   SETU - time to update the state.
	   TERR - time spent on travelling back and in the receive_reply method of the client interc.
	   RTT  - round trip time



	*/


	try{
	    java.io.RandomAccessFile table=new java.io.RandomAccessFile(experim_name,"rw"),value_file=new java.io.RandomAccessFile(value_file_name,"rw"),value_file_w_wait=new java.io.RandomAccessFile(value_file_w_wait_name,"rw");

	    table.writeBytes("Ret.Id.|                           Client Id.                      |Operation|Time\n");
	    table.writeBytes("---------------------------------------------------------------------------------\n\n");
	    int size,i,j;
	    /*String opers_cwpr[][]={"SCL","TTS","SUSFE","SUSWP","SUSREE","TUDRRP","TUFSR","TSGS","TUERR","ASP","TSR","TERR","TRE","TRO"}, opers_sr[]={"SCL","TTS","SUSFE","SUSWP","SUSREE","TUDRRP","TUFSR","TUERR","ASP","TSR","TERR"};*/
	    long wait,rtt;
	    ReqTraceStr stat_elem;
	    java.util.Set entrySet=req_trace_t.entrySet();
	    java.util.Iterator iter=entrySet.iterator();
	    java.util.Map.Entry entry;
	    java.util.Vector ret_cl_id_vector;
	    boolean possible_paths[],put;
	    int retention_id;
	    String client_id, oper_name;
	    long time_enter_send_request=0,time_exit_send_request=0,time_exit_receive_reply=0,time_enter_receive_request=0,time_start_wait_upd_oper=0,time_stop_wait_upd_oper=0,time_start_proc=0,time_start_lookup=0,time_stop_lookup=0,time_stop_proc_reply_there=0,time_start_tent_exec=0,time_start_get_state_b=0,time_start_invoke=0,time_start_get_state_a=0,time_stop_tent_exec=0,time_start_consensus=0,time_stop_consensus=0,time_start_restore_st=0,time_stop_restore_st=0,time_stop_proc_exc=0,time_set_state=0,time_set_update=0,time_stop_set_update=0,time_stop_proc_normal=0,interval,interval1,total_time_crt,total_time_w;
	    AvRecStr av_str;
	    java.util.Vector avg_vector, vectors[];
	    AvWaitRecStr av_w_str;
	    boolean print_time_e_s_r;
	    AvSliceRecStr av_sl_str;
	    boolean decr,wait_there,rtt_there;
	    System.out.println("The nr of requests is "+req_trace_t.size());
	    while(iter.hasNext())
		{
		    time_enter_send_request=0;
		    time_exit_send_request=0;
		    time_exit_receive_reply=0;
		    time_enter_receive_request=0;
		    time_start_wait_upd_oper=0;
		    time_stop_wait_upd_oper=0;
		    time_start_proc=0;
		    time_start_lookup=0;
		    time_stop_lookup=0;
		    time_stop_proc_reply_there=0;
		    time_start_tent_exec=0;
		    time_start_get_state_b=0;
		    time_start_invoke=0;
		    time_start_get_state_a=0;
		    time_stop_tent_exec=0;
		    time_start_consensus=0;
		    time_stop_consensus=0;
		    time_start_restore_st=0;
		    time_stop_restore_st=0;
		    time_stop_proc_exc=0;
		    time_set_state=0;
		    time_set_update=0;
		    time_stop_set_update=0;
		    time_stop_proc_normal=0;
		    interval=0;
		    interval1=0;

		    rtt_there=false;
		    wait_there=false;
		    table.writeBytes("\n");
		    entry=(java.util.Map.Entry)iter.next();
		    ret_cl_id_vector=(java.util.Vector)entry.getValue();
		    size=ret_cl_id_vector.size();
		    if(size==0) continue;

		    retention_id=((ReqTraceStr)ret_cl_id_vector.elementAt(0)).retention_id;
		    client_id=((ReqTraceStr)ret_cl_id_vector.elementAt(0)).client_id;

		    oper_name=((ReqTraceStr)ret_cl_id_vector.elementAt(0)).oper_name;
		    if(oper_name==null)
			continue;
		    vectors=(java.util.Vector[])times.get(oper_name);
		    if(vectors==null){
		      vectors=new java.util.Vector[19];
		      times.put(oper_name,vectors);
		      for(i=0;i<19;i++)
			  vectors[i]=new java.util.Vector();
		    }



		    av_str=null;
		    avg_vector=null;
		    try{
			avg_vector=(java.util.Vector)oper_av_t.get(oper_name);
		    }catch(java.lang.Exception e){e.printStackTrace(new java.io.PrintStream(System.out));}


		    put=false;
		    if(avg_vector==null)
			{
			    avg_vector=new java.util.Vector();
			    av_w_str=new AvWaitRecStr();
			    av_str=new AvRecStr();
			    put=true;
			    avg_vector.addElement(av_str);
			    avg_vector.addElement(av_w_str);
			}
		    else
			{
			    av_str=(AvRecStr)avg_vector.elementAt(0);
			    av_w_str=(AvWaitRecStr)avg_vector.elementAt(1);

			}
		    wait=0;
		    rtt=0;
		    //	if(av_str==null) {av_w_str=new AvWaitRecStr();av_str=new AvRecStr();put=true;}
		    if(retention_id==-1) continue;
		    total_time_crt=av_str.sum;
		    total_time_w=av_w_str.sum;
		    av_str.nr_of_calls++;
		    av_w_str.nr_of_calls++;
		    table.writeBytes(pad_nr(retention_id,NR_DIGITS_NR)+"|"+client_id+"|"+pad_str("",NR_CHARS_OPER)+"|"+oper_name+"\n\n");
		    //System.out.println("Size of vector is "+size);
		    //if(size>35) continue;
		    sort_vector_t(ret_cl_id_vector);
		    //System.out.println("Sorted vector of size "+size);
		    decr=false;
		    print_time_e_s_r=false;

		    for(i=0;i<size;i++)
			{
			    stat_elem=(ReqTraceStr)ret_cl_id_vector.elementAt(i);
			    switch(stat_elem.moment_code){
			    case universal.utils.Utils.ENTER_SEND_REQUEST:
				// for(j=0;j<possible_paths.length;j++)
				//possible_paths[j]=true;
				time_enter_send_request=stat_elem.time;
				break;
			    case universal.utils.Utils.EXIT_SEND_REQUEST:
				//for(j=0;j<possible_paths.length;j++)
				//possible_paths[j]=true;
				time_exit_send_request=stat_elem.time;
				if(time_enter_send_request!=0)
				    interval=time_exit_send_request-time_enter_send_request;
				else
				    interval=0;

				//av_str.sum+=interval;
				if(print_time_e_s_r)
				    System.out.println("time_exit_send_reques for "+oper_name+" is "+time_exit_send_request);
				av_sl_str=get_slice("SCL",avg_vector);
				if(interval>av_sl_str.max_value)
				    av_sl_str.max_value=interval;
				av_sl_str.sum+=interval;
				av_sl_str.sum_sq+=interval*interval;
				av_sl_str.nr_of_calls++;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("SCL",NR_CHARS_OPER)+"|"+interval+"\n");
				vectors[0].addElement(new Long(interval));
				break;
			    case universal.utils.Utils.ENTER_RECEIVE_REQUEST:
				//for(j=0;j<possible_paths.length;j++)
				//possible_paths[j]=true;
				time_enter_receive_request=stat_elem.time;
				if(time_exit_send_request!=0)
				    interval=time_enter_receive_request-time_exit_send_request;
				else{
				    System.out.println("TTS interval for "+oper_name+" is 0 and time_enter_rec_r is "+time_enter_receive_request);
				    print_time_e_s_r=true;
				    interval=0;
				}
				//av_str.sum+=interval;
				av_sl_str=get_slice("TTS",avg_vector);
				if(interval>av_sl_str.max_value)
				    av_sl_str.max_value=interval;
				av_sl_str.sum+=interval;
				av_sl_str.sum_sq+=interval*interval;
				av_sl_str.nr_of_calls++;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("TTS",NR_CHARS_OPER)+"|"+interval+"\n");
				vectors[1].addElement(new Long(interval));
				break;
			    case universal.utils.Utils.START_WAIT_UPD_OPER:
				//for(j=0;j<possible_paths.length;j++)
				//if(j!=0)
				// possible_paths[j]=false;
				time_start_wait_upd_oper=stat_elem.time;
				break;
			    case universal.utils.Utils.STOP_WAIT_UPD_OPER:
				//for(j=0;j<possible_paths.length;j++)
				//if(j!=1)
				//possible_paths[j]=false;
				time_stop_wait_upd_oper=stat_elem.time;
				if(time_start_wait_upd_oper!=0)
				    interval=time_stop_wait_upd_oper-time_start_wait_upd_oper;
				else
				    interval=0;
				av_sl_str=get_slice("TW",avg_vector);
				if(interval>av_sl_str.max_value)
				    av_sl_str.max_value=interval;
				av_sl_str.sum+=interval;
				av_sl_str.sum_sq+=interval*interval;
				av_sl_str.nr_of_calls++;
				av_w_str.sum+=interval;
				wait=interval;
				if(rtt_there){
				    av_str.sum_sq_without_wait+=(rtt-wait)*(rtt-wait);
				    av_str.sum_without_wait+=rtt-wait;
				    if(av_str.max_value_without_wait<rtt-wait)
					av_str.max_value_without_wait=rtt-wait;
				}
				wait_there=true;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("TW",NR_CHARS_OPER)+"|"+interval+"\n");
				vectors[2].addElement(new Long(interval));
				break;
			    case universal.utils.Utils.START_PROC:
				time_start_proc=stat_elem.time;

				break;
			    case universal.utils.Utils.START_LOOKUP:
				time_start_lookup=stat_elem.time;
				break;

			    case universal.utils.Utils.STOP_LOOKUP:
				time_stop_lookup=stat_elem.time;
				if(time_start_lookup!=0)
				    interval=time_stop_lookup-time_start_lookup;
				else
				    interval=0;
				av_sl_str=get_slice("TWRAC",avg_vector);
				if(interval>av_sl_str.max_value)
				    av_sl_str.max_value=interval;
				av_sl_str.sum+=interval;
				av_sl_str.sum_sq+=interval*interval;
				av_sl_str.nr_of_calls++;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("TWRAC",NR_CHARS_OPER)+"|"+interval+"\n");
				vectors[3].addElement(new Long(interval));
				/* TUSRMC - time until start record method call in cold passive*/
				break;
			    case universal.utils.Utils.STOP_PROC_REPLY_THERE:
				time_stop_proc_reply_there=stat_elem.time;
				if(time_start_proc!=0)
				    interval=time_stop_proc_reply_there-time_start_proc;
				else
				    interval=0;
				if(time_enter_receive_request!=0)
				    interval1=time_stop_proc_reply_there-time_enter_receive_request;
				//av_str.sum+=interval;
				av_sl_str=get_slice("TPRA",avg_vector);
				if(interval>av_sl_str.max_value)
				    av_sl_str.max_value=interval;
				av_sl_str.sum+=interval;
				av_sl_str.sum_sq+=interval*interval;
				av_sl_str.nr_of_calls++;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("TPRA",NR_CHARS_OPER)+"|"+interval+"\n");
				vectors[4].addElement(new Long(interval));
				av_sl_str=get_slice("SRA",avg_vector);
				if(interval1>av_sl_str.max_value)
				    av_sl_str.max_value=interval1;
				av_sl_str.sum+=interval1;
				av_sl_str.sum_sq+=interval1*interval1;
				av_sl_str.nr_of_calls++;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("SRA",NR_CHARS_OPER)+"|"+interval1+"\n");
				vectors[5].addElement(new Long(interval1));
				break;
			    case universal.utils.Utils.START_TENT_EXEC:
				// for(j=0;j<possible_paths.length;j++)
				//if(j!=3)
				//  possible_paths[j]=false;
				time_start_tent_exec=stat_elem.time;

				break;
			    case universal.utils.Utils.START_GET_STATE_B:
				//for(j=0;j<possible_paths.length;j++)
				//if(j!=4)
				//possible_paths[j]=false;
				time_start_get_state_b=stat_elem.time;

				break;
			    case universal.utils.Utils.START_INVOKE:
				time_start_invoke=stat_elem.time;
				if(time_start_get_state_b!=0)
				    interval=time_start_invoke-time_start_get_state_b;
				else
				    interval=0;
				av_sl_str=get_slice("TGSB",avg_vector);
				if(interval>av_sl_str.max_value)
				    av_sl_str.max_value=interval;
				av_sl_str.sum+=interval;
				av_sl_str.sum_sq+=interval*interval;
				av_sl_str.nr_of_calls++;
				// av_str.sum+=interval;//For now
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("TGSB",NR_CHARS_OPER)+"|"+interval+"\n");
				vectors[6].addElement(new Long(interval));
				break;
			    case universal.utils.Utils.START_GET_STATE_A:
				time_start_get_state_a=stat_elem.time;
				if(time_start_invoke!=0)
				    interval=time_start_get_state_a-time_start_invoke;
				else
				    interval=0;
				av_sl_str=get_slice("TER",avg_vector);
				if(interval>av_sl_str.max_value)
				    av_sl_str.max_value=interval;
				av_sl_str.sum+=interval;
				av_sl_str.sum_sq+=interval*interval;
				av_sl_str.nr_of_calls++;
				// av_str.sum+=interval;//For now
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("TER",NR_CHARS_OPER)+"|"+interval+"\n");
				vectors[7].addElement(new Long(interval));
				break;
			    case universal.utils.Utils.STOP_TENT_EXEC:
				time_stop_tent_exec=stat_elem.time;
				if(time_start_get_state_a!=0)
				    interval=time_stop_tent_exec-time_start_get_state_a;
				else
				    interval=0;
				av_sl_str=get_slice("TGSA",avg_vector);
				if(interval>av_sl_str.max_value)
				    av_sl_str.max_value=interval;
				av_sl_str.sum+=interval;
				av_sl_str.sum_sq+=interval*interval;
				av_sl_str.nr_of_calls++;
				//av_str.sum+=interval;//For now
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("TGSA",NR_CHARS_OPER)+"|"+interval+"\n");
				vectors[8].addElement(new Long(interval));
				break;
			    case universal.utils.Utils.START_CONSENSUS:
				time_start_consensus=stat_elem.time;
				break;
			    case universal.utils.Utils.STOP_CONSENSUS:
				time_stop_consensus=stat_elem.time;
				if(time_start_consensus!=0)
				    interval=time_stop_consensus-time_start_consensus;
				else
				    interval=0;
				av_sl_str=get_slice("TC",avg_vector);
				if(interval>av_sl_str.max_value)
				    av_sl_str.max_value=interval;
				av_sl_str.sum+=interval;
				av_sl_str.sum_sq+=interval*interval;
				av_sl_str.nr_of_calls++;
				//av_str.sum+=interval;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("TC",NR_CHARS_OPER)+"|"+interval+"\n");
				vectors[9].addElement(new Long(interval));
				/* TRR time to record reply (without broadcast in COLD_PASSIVE)*/
				break;
			    case universal.utils.Utils.START_RESTORE_ST:
				time_start_restore_st=stat_elem.time;

				break;
			    case universal.utils.Utils.STOP_RESTORE_ST:
				time_stop_restore_st=stat_elem.time;
				if(time_start_restore_st!=0)
				    interval=time_stop_restore_st-time_start_restore_st;
				else
				    interval=0;
				av_sl_str=get_slice("TRS",avg_vector);
				if(interval>av_sl_str.max_value)
				    av_sl_str.max_value=interval;
				av_sl_str.sum+=interval;
				av_sl_str.sum_sq+=interval*interval;
				av_sl_str.nr_of_calls++;
				//av_str.sum+=interval;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("TRS",NR_CHARS_OPER)+"|"+interval+"\n");
				vectors[10].addElement(new Long(interval));
				break;

			    case universal.utils.Utils.STOP_PROC_EXC:
				time_stop_proc_exc=stat_elem.time;
				if(time_start_proc!=0)
				    interval=time_stop_proc_exc-time_start_proc;
				else
				    interval=0;
				if(time_enter_receive_request!=0)
				    interval1=time_stop_proc_exc-time_enter_receive_request;
				//av_str.sum+=interval;
				av_sl_str=get_slice("TPSNL",avg_vector);
				if(interval>av_sl_str.max_value)
				    av_sl_str.max_value=interval;
				av_sl_str.sum+=interval;
				av_sl_str.sum_sq+=interval*interval;
				av_sl_str.nr_of_calls++;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("TPSNL",NR_CHARS_OPER)+"|"+interval+"\n");
				vectors[11].addElement(new Long(interval));
				av_sl_str=get_slice("SSNL",avg_vector);
				if(interval1>av_sl_str.max_value)
				    av_sl_str.max_value=interval1;
				av_sl_str.sum+=interval1;
				av_sl_str.sum_sq+=interval1*interval1;
				av_sl_str.nr_of_calls++;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("SSNL",NR_CHARS_OPER)+"|"+interval1+"\n");
				vectors[12].addElement(new Long(interval1));
				break;
			    case universal.utils.Utils.SET_STATE:
				time_set_state=stat_elem.time;
				break;
			    case universal.utils.Utils.SET_UPDATE:
				time_set_update=stat_elem.time;
				if(time_set_state!=0)
				    interval=time_set_update-time_set_state;
				else
				    interval=0;
				av_sl_str=get_slice("SETS",avg_vector);
				if(interval>av_sl_str.max_value)
				    av_sl_str.max_value=interval;
				av_sl_str.sum+=interval;
				av_sl_str.sum_sq+=interval*interval;
				av_sl_str.nr_of_calls++;
				//av_str.sum+=interval;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("SETS",NR_CHARS_OPER)+"|"+interval+"\n");
				vectors[13].addElement(new Long(interval));
				break;
			    case universal.utils.Utils.STOP_SET_UPDATE:
				time_stop_set_update=stat_elem.time;
				if(time_set_update!=0)
				    interval=time_stop_set_update-time_set_state;
				else
				    interval=0;
				av_sl_str=get_slice("SETU",avg_vector);
				if(interval>av_sl_str.max_value)
				    av_sl_str.max_value=interval;
				av_sl_str.sum+=interval;
				av_sl_str.sum_sq+=interval*interval;
				av_sl_str.nr_of_calls++;
				//av_str.sum+=interval;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("SETU",NR_CHARS_OPER)+"|"+interval+"\n");
				vectors[14].addElement(new Long(interval));
				break;
			    case universal.utils.Utils.STOP_PROC_NORMAL:
				time_stop_proc_normal=stat_elem.time;
				if(time_start_proc!=0)
				    interval=time_stop_proc_normal-time_start_proc;
				else
				    interval=0;
				if(time_enter_receive_request!=0)
				    interval1=time_stop_proc_normal-time_enter_receive_request;
				//av_str.sum+=interval;
				av_sl_str=get_slice("TPSL",avg_vector);
				if(interval>av_sl_str.max_value)
				    av_sl_str.max_value=interval;
				av_sl_str.sum+=interval;
				av_sl_str.sum_sq+=interval*interval;
				av_sl_str.nr_of_calls++;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("TPSL",NR_CHARS_OPER)+"|"+interval+"\n");
				vectors[15].addElement(new Long(interval));
				av_sl_str=get_slice("SSL",avg_vector);
				if(interval1>av_sl_str.max_value)
				    av_sl_str.max_value=interval1;
				av_sl_str.sum+=interval1;
				av_sl_str.sum_sq+=interval1*interval1;
				av_sl_str.nr_of_calls++;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("SSL",NR_CHARS_OPER)+"|"+interval1+"\n");
				vectors[16].addElement(new Long(interval1));
				break;
			    case universal.utils.Utils.EXIT_RECEIVE_REPLY:
				time_exit_receive_reply=stat_elem.time;
				if(time_stop_proc_reply_there!=0)
				    interval=time_exit_receive_reply-time_stop_proc_reply_there;
				else
				    {
					if(time_stop_proc_exc!=0)
					    interval=time_exit_receive_reply-time_stop_proc_exc;
					else
					    {
						if(time_stop_proc_normal!=0)
						    interval=time_exit_receive_reply-time_stop_proc_normal;
						else
						    interval=0;
					    }
				    }
				if(time_enter_send_request!=0)
				    {
					interval1=time_exit_receive_reply-time_enter_send_request;
				    }
				else
				    interval1=0;
				av_sl_str=get_slice("TERR",avg_vector);
				if(interval>av_sl_str.max_value)
				    av_sl_str.max_value=interval;
				av_sl_str.sum+=interval;
				av_sl_str.sum_sq+=interval*interval;
				av_sl_str.nr_of_calls++;
				//av_str.sum+=interval;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("TERR",NR_CHARS_OPER)+"|"+interval+"\n");
				vectors[17].addElement(new Long(interval));
				av_sl_str=get_slice("RTT",avg_vector);
				if(interval1>av_sl_str.max_value)
				    av_sl_str.max_value=interval1;
				av_sl_str.sum+=interval1;
				av_sl_str.sum_sq+=interval1*interval1;
				av_sl_str.nr_of_calls++;
				//av_str.sum+=interval;
				table.writeBytes(spaces(NR_DIGITS_NR)+"|"+spaces(client_id.length())+"|"+pad_str("RTT",NR_CHARS_OPER)+"|"+interval1+"\n");
				av_str.sum+=interval1;
				av_str.sum_sq+=interval1*interval1;
				//if(wait==0) System.out.println("OOps wait is 0 for "+oper_name);
				if(wait_there){
				    //if(wait==0) System.out.println("Oops wait is 0 for "+oper_name+" "+retention_id);
				    av_str.sum_sq_without_wait+=(interval1-wait)*(interval1-wait);
				    av_str.sum_without_wait+=interval1-wait;
				//    av_str.values.addElement(new Long(interval1-wait));
				    if(av_str.max_value_without_wait<interval1-wait)
					av_str.max_value_without_wait=interval1-wait;
				}
				rtt_there=true;
				rtt=interval1;
				if(interval1>av_str.max_value)
				    av_str.max_value=interval1;
				//value_vect.addElement(new Long(rtt));
				//value_vect.addElement(new Long(rtt-wait));
				vectors[18].addElement(new Long(interval1));
				break;
			    }

			}
		    if(put) {
			oper_av_t.put(oper_name,avg_vector);
				//  oper_av_w_t.put(oper_name,av_w_str);
		    }

		}
	    table.close();
	    System.out.println("Table close");

	}catch(java.lang.Exception e){e.printStackTrace();}
    }


    String names[]={"SCL","TTS","TW","TWRAC","TPRA","SRA","TGSB","TER","TGSA","TC","TRS","TPSNL","SSNL","SETS","SETU","TPSL","SSL","TERR","RTT"};

    private void generate_files(String directory){
	java.util.Set entries;
	java.util.Iterator e_iter;
	java.util.Map.Entry elem;
	java.util.Vector[] vectors;
	String op_n;
	int i,j,l;
	try{
	    entries=times.entrySet();
	    e_iter=entries.iterator();
	    java.io.RandomAccessFile to_write;
	    while(e_iter.hasNext()){
		elem=(java.util.Map.Entry)e_iter.next();
		op_n=(String)elem.getKey();
		vectors=(java.util.Vector[])elem.getValue();
		for(i=0;i<vectors.length;i++){
		    l=vectors[i].size();
		    if(l>0){
			to_write=new java.io.RandomAccessFile(directory+"/"+op_n+"_"+names[i],"rw");
			for(j=0;j<l;j++)
			    to_write.writeBytes(vectors[i].elementAt(j)+"\n");

		    }
		}
	    }
	}catch(java.lang.Exception e){}

    }


    private  void printAvg(java.util.Hashtable avg_t,String file,boolean correct){
	java.io.RandomAccessFile avg_f=null;
	try{
	    avg_f=new java.io.RandomAccessFile(file,"rw");
	}catch(java.lang.Exception e){avg_f=null;}

	java.util.Set entries;
	java.util.Iterator e_iter;
	java.util.Map.Entry elem;
	java.util.Vector avg_vector;
	String op_n;
	AvRecStr av_r;
	AvWaitRecStr av_w_r;
	AvSliceRecStr av_sl_str;
	int nr_of_calls,l,i;
	float av_s,av_w,av_sl,std_dev_w,av_s_absolute;
	double st_dev_sl,st_dev_s;
	entries=avg_t.entrySet();
	e_iter=entries.iterator();
	String sample_values="";
	while(e_iter.hasNext()){
	    elem=(java.util.Map.Entry)e_iter.next();
	    op_n=(String)elem.getKey();
	    avg_vector=(java.util.Vector)elem.getValue();
	    av_r=(AvRecStr)avg_vector.elementAt(0);
	    av_w_r=(AvWaitRecStr)avg_vector.elementAt(1);
	    System.out.println(op_n+" Roundtrip time max value is "+av_r.max_value);
	    av_s_absolute=(float)(((double)av_r.sum)/((double)av_r.nr_of_calls));
	    if(av_r.nr_of_calls>1){
		if(correct){
		    av_s_absolute=(float)(((double)(av_r.sum-av_r.max_value))/((double)(av_r.nr_of_calls-1)));
		    av_r.sum=av_r.sum_without_wait-av_r.max_value_without_wait;
		    av_r.sum_sq=av_r.sum_sq_without_wait-av_r.max_value_without_wait*av_r.max_value_without_wait;
		}
		else{
		    av_r.sum-=av_r.max_value;
		    av_r.sum_sq-=av_r.max_value*av_r.max_value;
		}
		av_r.nr_of_calls--;
	    }
	    //sample_values+="The values for "+op_n+" are: ";
	    /*for(i=0;i<av_r.values.size();i++)
	      sample_values+=av_r.values.elementAt(i)+" ";*/
	    av_s=(float)(((double)av_r.sum)/((double)av_r.nr_of_calls));
	    if(!correct)
		av_s_absolute=av_s;
	    st_dev_s=((double)(av_r.sum_sq)/((double)av_r.nr_of_calls))-(double)(av_s*av_s);
	    if(st_dev_s>0)
		st_dev_s=Math.sqrt(st_dev_s);
	    else
		st_dev_s=-1;
	    if(avg_f!=null)
		try{
		    avg_f.writeBytes(op_n+" "+av_r.nr_of_calls+" "+av_s+" st. dev. "+st_dev_s+"\n");
		    //avg_f.writeBytes(sample_values);
		}catch(java.io.IOException e){}
	    else
		System.out.println(op_n+" has an average time of "+av_s+" for a number of "+av_r.nr_of_calls+" calls");

	    av_w=(float)(((double)av_w_r.sum)/((double)av_w_r.nr_of_calls));
	    if(avg_f!=null)
		try{
		    if(av_s!=0)
			avg_f.writeBytes(op_n+" "+av_w_r.nr_of_calls+" "+((float)(100.0*(av_w/av_s_absolute)))+"% "+av_w+"\n");
		}catch(java.io.IOException e){}
	    else
		System.out.println(op_n+" has an average wait time of "+av_w+" for a number of "+av_w_r.nr_of_calls+" calls");
	    l=avg_vector.size();
	    for(i=2;i<l;i++){
		av_sl_str=(AvSliceRecStr)avg_vector.elementAt(i);
		System.out.println(op_n+" "+av_sl_str.oper+" max values is "+av_sl_str.max_value);
		if(av_sl_str.nr_of_calls>1){
		    av_sl_str.sum-=av_sl_str.max_value;
		    av_sl_str.sum_sq-=av_sl_str.max_value*av_sl_str.max_value;
		    av_sl_str.nr_of_calls--;
		}
		av_sl=(float)(((double)av_sl_str.sum)/((double)av_sl_str.nr_of_calls));
		st_dev_sl=((double)(av_sl_str.sum_sq)/((double)av_sl_str.nr_of_calls))-(double)(((double)(av_sl_str.sum*av_sl_str.sum))/((double)(av_sl_str.nr_of_calls*av_sl_str.nr_of_calls)));
		if(st_dev_sl>0)
		    st_dev_sl=Math.sqrt(st_dev_sl);
		else
		    st_dev_sl=-1;
		if(avg_f!=null)
		    try{
			if(av_s!=0)
			    avg_f.writeBytes(op_n+" "+pad_str(av_sl_str.oper,NR_CHARS_OPER)+" "+av_sl_str.nr_of_calls+" "+((float)(100.0*(av_sl/av_s)))+"% "+av_sl+" st. dev. "+st_dev_sl+"\n");
		    }catch(java.io.IOException e){}
		else
		    System.out.println(op_n+" spent with operation "+av_sl_str.oper+" "+av_sl+" for a number of "+av_r.nr_of_calls+" calls");
	    }
	}
    }




    public static void main(String args[]){
	Compiler compiler=new Compiler();
	boolean do_correct;
	if(args[0].charAt(0)=='0') do_correct=false;
	else
	    do_correct=true;


	String directory = args[1];
	java.io.File file=new java.io.File(args[1]);
	java.io.File[] its_files=file.listFiles();
	int i;
	for(i=0;i<its_files.length;i++){
	    if(its_files[i].getName().indexOf("%%")!=-1){
		compiler.parse_req_trace(its_files[i].getName());
	    }

	}
	compiler.compute_req_times(args[2],args[4],args[5]);

	compiler.generate_files(args[3]);

    }
}
